@@ -250,6 +250,15 @@ static bool canTransformMesh(const Mesh& mesh)
 	return true;
 }
 
+static bool isExtensionSupported(const ExtensionInfo* extensions, size_t count, const char* name)
+{
+	for (size_t i = 0; i < count; ++i)
+		if (strcmp(extensions[i].name, name) == 0)
+			return true;
+
+	return false;
+}
+
 static void process(cgltf_data* data, const char* input_path, const char* output_path, const char* report_path, std::vector<Mesh>& meshes, std::vector<Animation>& animations, const Settings& settings, std::string& json, std::string& bin, std::string& fallback, size_t& fallback_size)
 {
 	if (settings.verbose)
@@ -272,7 +281,10 @@ static void process(cgltf_data* data, const char* input_path, const char* output
 	for (size_t i = 0; i < meshes.size(); ++i)
 	{
 		Mesh& mesh = meshes[i];
-		assert(mesh.instances.empty());
+
+		// mesh is already instanced, skip
+		if (!mesh.instances.empty())
+			continue;
 
 		// mesh is already world space, skip
 		if (mesh.nodes.empty())
@@ -352,8 +364,9 @@ static void process(cgltf_data* data, const char* input_path, const char* output
 		{
 			MaterialInfo vi = materials[mesh.variants[j].material - data->materials];
 
-			mi.needsTangents |= vi.needsTangents;
-			mi.textureSetMask |= vi.textureSetMask;
+			mi.needs_tangents |= vi.needs_tangents;
+			mi.texture_set_mask |= vi.texture_set_mask;
+			mi.unlit &= vi.unlit;
 		}
 
 		filterStreams(mesh, mi);
@@ -377,7 +390,7 @@ static void process(cgltf_data* data, const char* input_path, const char* output
 		{
 			Mesh kinds = {};
 			Mesh loops = {};
-			debugSimplify(mesh, kinds, loops, settings.simplify_debug);
+			debugSimplify(mesh, kinds, loops, settings.simplify_debug, settings.simplify_attributes);
 			debug_meshes.push_back(kinds);
 			debug_meshes.push_back(loops);
 		}
@@ -385,10 +398,8 @@ static void process(cgltf_data* data, const char* input_path, const char* output
 		if (settings.meshlet_debug > 0)
 		{
 			Mesh meshlets = {};
-			Mesh bounds = {};
-			debugMeshlets(mesh, meshlets, bounds, settings.meshlet_debug, /* scan= */ false);
+			debugMeshlets(mesh, meshlets, settings.meshlet_debug, /* scan= */ false);
 			debug_meshes.push_back(meshlets);
-			debug_meshes.push_back(bounds);
 		}
 	}
 #endif
@@ -438,9 +449,12 @@ static void process(cgltf_data* data, const char* input_path, const char* output
 	bool ext_emissive_strength = false;
 	bool ext_iridescence = false;
 	bool ext_anisotropy = false;
+	bool ext_dispersion = false;
 	bool ext_unlit = false;
 	bool ext_instancing = false;
 	bool ext_texture_transform = false;
+	bool ext_texture_basisu = false;
+	bool ext_texture_webp = false;
 
 	size_t accr_offset = 0;
 	size_t node_offset = 0;
@@ -473,22 +487,15 @@ static void process(cgltf_data* data, const char* input_path, const char* output
 	{
 		const cgltf_image& image = data->images[i];
 
+		std::string* encoded = (encoded_images.size() && !encoded_images[i].empty()) ? &encoded_images[i] : NULL;
+
 		comma(json_images);
 		append(json_images, "{");
-		if (encoded_images.size() && !encoded_images[i].empty())
-		{
-			if (encoded_images[i].compare(0, 5, "error") == 0)
-				fprintf(stderr, "Warning: unable to encode image %d (%s), skipping (%s)\n", int(i), image.uri ? image.uri : "?", encoded_images[i].c_str());
-			else
-				writeEncodedImage(json_images, views, image, encoded_images[i], images[i], output_path, settings);
-
-			encoded_images[i] = std::string(); // reclaim memory early
-		}
-		else
-		{
-			writeImage(json_images, views, image, images[i], i, input_path, output_path, settings);
-		}
+		writeImage(json_images, views, image, images[i], encoded, i, input_path, output_path, settings);
 		append(json_images, "}");
+
+		if (encoded)
+			*encoded = std::string(); // reclaim memory early
 	}
 
 	for (size_t i = 0; i < data->textures_count; ++i)
@@ -505,6 +512,8 @@ static void process(cgltf_data* data, const char* input_path, const char* output
 
 		assert(textures[i].remap == int(texture_offset));
 		texture_offset++;
+		ext_texture_basisu = ext_texture_basisu || texture.has_basisu;
+		ext_texture_webp = ext_texture_webp || texture.has_webp;
 	}
 
 	for (size_t i = 0; i < data->materials_count; ++i)
@@ -536,8 +545,9 @@ static void process(cgltf_data* data, const char* input_path, const char* output
 		ext_emissive_strength = ext_emissive_strength || material.has_emissive_strength;
 		ext_iridescence = ext_iridescence || material.has_iridescence;
 		ext_anisotropy = ext_anisotropy || material.has_anisotropy;
+		ext_dispersion = ext_dispersion || material.has_dispersion;
 		ext_unlit = ext_unlit || material.unlit;
-		ext_texture_transform = ext_texture_transform || mi.usesTextureTransform;
+		ext_texture_transform = ext_texture_transform || mi.uses_texture_transform;
 	}
 
 	for (size_t i = 0; i < meshes.size(); ++i)
@@ -573,7 +583,7 @@ static void process(cgltf_data* data, const char* input_path, const char* output
 			if (prim.type != cgltf_primitive_type_triangles)
 			{
 				append(json_meshes, ",\"mode\":");
-				append(json_meshes, size_t(prim.type));
+				append(json_meshes, size_t(prim.type - cgltf_primitive_type_points));
 			}
 			if (mesh.targets)
 			{
@@ -657,9 +667,6 @@ static void process(cgltf_data* data, const char* input_path, const char* output
 
 		append(json_meshes, "}");
 
-		assert(mesh.nodes.empty() || mesh.instances.empty());
-		ext_instancing = ext_instancing || !mesh.instances.empty();
-
 		if (mesh.nodes.size())
 		{
 			for (size_t j = 0; j < mesh.nodes.size(); ++j)
@@ -668,7 +675,7 @@ static void process(cgltf_data* data, const char* input_path, const char* output
 				assert(ni.keep);
 
 				// if we don't use position quantization, prefer attaching the mesh to its node directly
-				if (!ni.has_mesh && (!settings.quantize || settings.pos_float))
+				if (!ni.has_mesh && (!settings.quantize || settings.pos_float || (qp.offset[0] == 0.f && qp.offset[1] == 0.f && qp.offset[2] == 0 && qp.node_scale == 1.f)))
 				{
 					ni.has_mesh = true;
 					ni.mesh_index = mesh_offset;
@@ -684,7 +691,8 @@ static void process(cgltf_data* data, const char* input_path, const char* output
 				}
 			}
 		}
-		else if (mesh.instances.size())
+
+		if (mesh.instances.size())
 		{
 			assert(mesh.scene >= 0);
 			comma(json_roots[mesh.scene]);
@@ -697,7 +705,8 @@ static void process(cgltf_data* data, const char* input_path, const char* output
 
 			node_offset++;
 		}
-		else
+
+		if (mesh.nodes.empty() && mesh.instances.empty())
 		{
 			assert(mesh.scene >= 0);
 			comma(json_roots[mesh.scene]);
@@ -709,6 +718,7 @@ static void process(cgltf_data* data, const char* input_path, const char* output
 		}
 
 		mesh_offset++;
+		ext_instancing = ext_instancing || !mesh.instances.empty();
 
 		// skip all meshes that we've written in this iteration
 		assert(pi > i);
@@ -825,13 +835,23 @@ static void process(cgltf_data* data, const char* input_path, const char* output
 	    {"KHR_materials_emissive_strength", ext_emissive_strength, false},
 	    {"KHR_materials_iridescence", ext_iridescence, false},
 	    {"KHR_materials_anisotropy", ext_anisotropy, false},
+	    {"KHR_materials_dispersion", ext_dispersion, false},
 	    {"KHR_materials_unlit", ext_unlit, false},
 	    {"KHR_materials_variants", data->variants_count > 0, false},
 	    {"KHR_lights_punctual", data->lights_count > 0, false},
-	    {"KHR_texture_basisu", !json_textures.empty() && settings.texture_ktx2, true},
+	    {"KHR_texture_basisu", (!json_textures.empty() && settings.texture_ktx2) || ext_texture_basisu, true},
+	    {"EXT_texture_webp", ext_texture_webp, true},
 	    {"EXT_mesh_gpu_instancing", ext_instancing, true},
 	};
 
+	for (size_t i = 0; i < data->extensions_required_count; ++i)
+	{
+		const char* ext = data->extensions_required[i];
+
+		if (!isExtensionSupported(extensions, sizeof(extensions) / sizeof(extensions[0]), ext))
+			fprintf(stderr, "Warning: required extension %s is not supported and will be skipped\n", ext);
+	}
+
 	writeExtensions(json, extensions, sizeof(extensions) / sizeof(extensions[0]));
 
 	std::string json_views;
@@ -1167,7 +1187,7 @@ Settings defaults()
 template <typename T>
 T clamp(T v, T min, T max)
 {
-	return v < min ? min : v > max ? max : v;
+	return v < min ? min : (v > max ? max : v);
 }
 
 unsigned int textureMask(const char* arg)
@@ -1248,14 +1268,14 @@ int main(int argc, char** argv)
 		{
 			settings.pos_float = true;
 		}
-		else if (strcmp(arg, "-vtn") == 0)
-		{
-			settings.tex_float = false;
-		}
 		else if (strcmp(arg, "-vtf") == 0)
 		{
 			settings.tex_float = true;
 		}
+		else if (strcmp(arg, "-vnf") == 0)
+		{
+			settings.nrm_float = true;
+		}
 		else if (strcmp(arg, "-at") == 0 && i + 1 < argc && isdigit(argv[i + 1][0]))
 		{
 			settings.trn_bits = clamp(atoi(argv[++i]), 1, 24);
@@ -1308,6 +1328,10 @@ int main(int argc, char** argv)
 		{
 			settings.simplify_lock_borders = true;
 		}
+		else if (strcmp(arg, "-sv") == 0)
+		{
+			settings.simplify_attributes = true;
+		}
 #ifndef NDEBUG
 		else if (strcmp(arg, "-sd") == 0 && i + 1 < argc && isdigit(argv[i + 1][0]))
 		{
@@ -1493,19 +1517,20 @@ int main(int argc, char** argv)
 			fprintf(stderr, "\nSimplification:\n");
 			fprintf(stderr, "\t-si R: simplify meshes targeting triangle/point count ratio R (default: 1; R should be between 0 and 1)\n");
 			fprintf(stderr, "\t-sa: aggressively simplify to the target ratio disregarding quality\n");
+			fprintf(stderr, "\t-sv: take vertex attributes into account when simplifying meshes\n");
 			fprintf(stderr, "\t-slb: lock border vertices during simplification to avoid gaps on connected meshes\n");
-			fprintf(stderr, "\nVertices:\n");
+			fprintf(stderr, "\nVertex precision:\n");
 			fprintf(stderr, "\t-vp N: use N-bit quantization for positions (default: 14; N should be between 1 and 16)\n");
 			fprintf(stderr, "\t-vt N: use N-bit quantization for texture coordinates (default: 12; N should be between 1 and 16)\n");
-			fprintf(stderr, "\t-vn N: use N-bit quantization for normals and tangents (default: 8; N should be between 1 and 16)\n");
+			fprintf(stderr, "\t-vn N: use N-bit quantization for normals (default: 8; N should be between 1 and 16) and tangents (up to 8-bit)\n");
 			fprintf(stderr, "\t-vc N: use N-bit quantization for colors (default: 8; N should be between 1 and 16)\n");
 			fprintf(stderr, "\nVertex positions:\n");
 			fprintf(stderr, "\t-vpi: use integer attributes for positions (default)\n");
 			fprintf(stderr, "\t-vpn: use normalized attributes for positions\n");
 			fprintf(stderr, "\t-vpf: use floating point attributes for positions\n");
-			fprintf(stderr, "\nTexture coordinates:\n");
-			fprintf(stderr, "\t-vtn: use normalized attributes for texture coordinates (default)\n");
+			fprintf(stderr, "\nVertex attributes:\n");
 			fprintf(stderr, "\t-vtf: use floating point attributes for texture coordinates\n");
+			fprintf(stderr, "\t-vnf: use floating point attributes for normals\n");
 			fprintf(stderr, "\nAnimations:\n");
 			fprintf(stderr, "\t-at N: use N-bit quantization for translations (default: 16; N should be between 1 and 24)\n");
 			fprintf(stderr, "\t-ar N: use N-bit quantization for rotations (default: 12; N should be between 4 and 16)\n");
@@ -1581,9 +1606,9 @@ int main(int argc, char** argv)
 		return 1;
 	}
 
-	if (settings.fallback && (settings.pos_float || settings.tex_float))
+	if (settings.fallback && (settings.pos_float || settings.tex_float || settings.nrm_float))
 	{
-		fprintf(stderr, "Option -cf can not be used together with -vpf or -tpf\n");
+		fprintf(stderr, "Option -cf can not be used together with -vpf, -vtf or -vnf\n");
 		return 1;
 	}
 
