@@ -60,7 +60,7 @@ void appendJson(std::string& s, const char* data)
 			break;
 
 		case Quoted:
-			state = (ch == '"') ? None : (ch == '\\') ? Escape : Quoted;
+			state = (ch == '"') ? None : (ch == '\\' ? Escape : Quoted);
 			break;
 
 		case Escape:
