@@ -405,11 +405,11 @@ void filterStreams(Mesh& mesh, const MaterialInfo& mi)
 
 		if (stream.target)
 		{
-			morph_normal = morph_normal || (stream.type == cgltf_attribute_type_normal && !isConstant(stream.data, { 0, 0, 0, 0 }));
-			morph_tangent = morph_tangent || (stream.type == cgltf_attribute_type_tangent && !isConstant(stream.data, { 0, 0, 0, 0 }));
+			morph_normal = morph_normal || (stream.type == cgltf_attribute_type_normal && !isConstant(stream.data, {0, 0, 0, 0}));
+			morph_tangent = morph_tangent || (stream.type == cgltf_attribute_type_tangent && !isConstant(stream.data, {0, 0, 0, 0}));
 		}
 
-		if (stream.type == cgltf_attribute_type_texcoord && stream.index < 32 && (mi.textureSetMask & (1u << stream.index)) != 0)
+		if (stream.type == cgltf_attribute_type_texcoord && stream.index < 32 && (mi.texture_set_mask & (1u << stream.index)) != 0)
 		{
 			keep_texture_set = std::max(keep_texture_set, stream.index);
 		}
@@ -424,13 +424,16 @@ void filterStreams(Mesh& mesh, const MaterialInfo& mi)
 		if (stream.type == cgltf_attribute_type_texcoord && stream.index > keep_texture_set)
 			continue;
 
-		if (stream.type == cgltf_attribute_type_tangent && !mi.needsTangents)
+		if (stream.type == cgltf_attribute_type_normal && mi.unlit)
+			continue;
+
+		if (stream.type == cgltf_attribute_type_tangent && !mi.needs_tangents)
 			continue;
 
 		if ((stream.type == cgltf_attribute_type_joints || stream.type == cgltf_attribute_type_weights) && !mesh.skin)
 			continue;
 
-		if (stream.type == cgltf_attribute_type_color && isConstant(stream.data, { 1, 1, 1, 1 }))
+		if (stream.type == cgltf_attribute_type_color && isConstant(stream.data, {1, 1, 1, 1}))
 			continue;
 
 		if (stream.target && stream.type == cgltf_attribute_type_normal && !morph_normal)
@@ -524,7 +527,7 @@ static Stream* getStream(Mesh& mesh, cgltf_attribute_type type, int index = 0)
 	return NULL;
 }
 
-static void simplifyMesh(Mesh& mesh, float threshold, bool aggressive, bool lock_borders)
+static void simplifyMesh(Mesh& mesh, float threshold, bool attributes, bool aggressive, bool lock_borders)
 {
 	assert(mesh.type == cgltf_primitive_type_triangles);
 
@@ -542,11 +545,16 @@ static void simplifyMesh(Mesh& mesh, float threshold, bool aggressive, bool lock
 	float target_error_aggressive = 1e-1f;
 	unsigned int options = lock_borders ? meshopt_SimplifyLockBorder : 0;
 
-	if (target_index_count < 1)
-		return;
+	const Stream* attr = getStream(mesh, cgltf_attribute_type_color);
+	attr = attr ? attr : getStream(mesh, cgltf_attribute_type_normal);
+
+	const float attrw[3] = {1e-2f, 1e-2f, 1e-2f};
 
 	std::vector<unsigned int> indices(mesh.indices.size());
-	indices.resize(meshopt_simplify(&indices[0], &mesh.indices[0], mesh.indices.size(), positions->data[0].f, vertex_count, sizeof(Attr), target_index_count, target_error, options));
+	if (attributes && attr)
+		indices.resize(meshopt_simplifyWithAttributes(&indices[0], &mesh.indices[0], mesh.indices.size(), positions->data[0].f, vertex_count, sizeof(Attr), attr->data[0].f, sizeof(Attr), attrw, 3, NULL, target_index_count, target_error, options));
+	else
+		indices.resize(meshopt_simplify(&indices[0], &mesh.indices[0], mesh.indices.size(), positions->data[0].f, vertex_count, sizeof(Attr), target_index_count, target_error, options));
 	mesh.indices.swap(indices);
 
 	// Note: if the simplifier got stuck, we can try to reindex without normals/tangents and retry
@@ -702,13 +710,11 @@ static void simplifyPointMesh(Mesh& mesh, float threshold)
 
 	size_t target_vertex_count = size_t(double(vertex_count) * threshold);
 
-	if (target_vertex_count < 1)
-		return;
-
 	const float color_weight = 1e-2f;
 
 	std::vector<unsigned int> indices(target_vertex_count);
-	indices.resize(meshopt_simplifyPoints(&indices[0], positions->data[0].f, vertex_count, sizeof(Attr), colors ? colors->data[0].f : NULL, sizeof(Attr), color_weight, target_vertex_count));
+	if (target_vertex_count)
+		indices.resize(meshopt_simplifyPoints(&indices[0], positions->data[0].f, vertex_count, sizeof(Attr), colors ? colors->data[0].f : NULL, sizeof(Attr), color_weight, target_vertex_count));
 
 	std::vector<Attr> scratch(indices.size());
 
@@ -768,7 +774,7 @@ void processMesh(Mesh& mesh, const Settings& settings)
 		reindexMesh(mesh);
 		filterTriangles(mesh);
 		if (settings.simplify_threshold < 1)
-			simplifyMesh(mesh, settings.simplify_threshold, settings.simplify_aggressive, settings.simplify_lock_borders);
+			simplifyMesh(mesh, settings.simplify_threshold, settings.simplify_attributes, settings.simplify_aggressive, settings.simplify_lock_borders);
 		optimizeMesh(mesh, settings.compressmore);
 		break;
 
@@ -782,7 +788,7 @@ extern MESHOPTIMIZER_API unsigned char* meshopt_simplifyDebugKind;
 extern MESHOPTIMIZER_API unsigned int* meshopt_simplifyDebugLoop;
 extern MESHOPTIMIZER_API unsigned int* meshopt_simplifyDebugLoopBack;
 
-void debugSimplify(const Mesh& source, Mesh& kinds, Mesh& loops, float ratio)
+void debugSimplify(const Mesh& source, Mesh& kinds, Mesh& loops, float ratio, bool attributes)
 {
 	Mesh mesh = source;
 	assert(mesh.type == cgltf_primitive_type_triangles);
@@ -805,7 +811,7 @@ void debugSimplify(const Mesh& source, Mesh& kinds, Mesh& loops, float ratio)
 	meshopt_simplifyDebugLoop = &loop[0];
 	meshopt_simplifyDebugLoopBack = &loopback[0];
 
-	simplifyMesh(mesh, ratio, /* aggressive= */ false, /* lock_borders= */ false);
+	simplifyMesh(mesh, ratio, attributes, /* aggressive= */ false, /* lock_borders= */ false);
 
 	meshopt_simplifyDebugKind = NULL;
 	meshopt_simplifyDebugLoop = NULL;
@@ -878,7 +884,7 @@ void debugSimplify(const Mesh& source, Mesh& kinds, Mesh& loops, float ratio)
 		}
 }
 
-void debugMeshlets(const Mesh& source, Mesh& meshlets, Mesh& bounds, int max_vertices, bool scan)
+void debugMeshlets(const Mesh& source, Mesh& meshlets, int max_vertices, bool scan)
 {
 	Mesh mesh = source;
 	assert(mesh.type == cgltf_primitive_type_triangles);
@@ -941,66 +947,5 @@ void debugMeshlets(const Mesh& source, Mesh& meshlets, Mesh& bounds, int max_ver
 	meshlets.type = cgltf_primitive_type_triangles;
 	meshlets.streams.push_back(mv);
 	meshlets.streams.push_back(mc);
-
-	// generate bounds meshes, using a sphere per meshlet
-	bounds.nodes = mesh.nodes;
-
-	Stream bv = {cgltf_attribute_type_position};
-	Stream bc = {cgltf_attribute_type_color};
-
-	for (size_t i = 0; i < ml.size(); ++i)
-	{
-		const meshopt_Meshlet& m = ml[i];
-
-		meshopt_Bounds mb = meshopt_computeMeshletBounds(&mlv[m.vertex_offset], &mlt[m.triangle_offset], m.triangle_count, positions->data[0].f, positions->data.size(), sizeof(Attr));
-
-		unsigned int h = unsigned(i);
-		h ^= h >> 13;
-		h *= 0x5bd1e995;
-		h ^= h >> 15;
-
-		Attr c = {{float(h & 0xff) / 255.f, float((h >> 8) & 0xff) / 255.f, float((h >> 16) & 0xff) / 255.f, 0.1f}};
-
-		unsigned int offset = unsigned(bv.data.size());
-
-		const int N = 10;
-
-		for (int y = 0; y <= N; ++y)
-		{
-			float u = (y == N) ? 0 : float(y) / N * 2 * 3.1415926f;
-			float sinu = sinf(u), cosu = cosf(u);
-
-			for (int x = 0; x <= N; ++x)
-			{
-				float v = float(x) / N * 3.1415926f;
-				float sinv = sinf(v), cosv = cosf(v);
-
-				float fx = sinv * cosu;
-				float fy = sinv * sinu;
-				float fz = cosv;
-
-				Attr p = {{mb.center[0] + mb.radius * fx, mb.center[1] + mb.radius * fy, mb.center[2] + mb.radius * fz, 1.f}};
-
-				bv.data.push_back(p);
-				bc.data.push_back(c);
-			}
-		}
-
-		for (int y = 0; y < N; ++y)
-			for (int x = 0; x < N; ++x)
-			{
-				bounds.indices.push_back(offset + (N + 1) * (y + 0) + (x + 0));
-				bounds.indices.push_back(offset + (N + 1) * (y + 0) + (x + 1));
-				bounds.indices.push_back(offset + (N + 1) * (y + 1) + (x + 0));
-
-				bounds.indices.push_back(offset + (N + 1) * (y + 1) + (x + 0));
-				bounds.indices.push_back(offset + (N + 1) * (y + 0) + (x + 1));
-				bounds.indices.push_back(offset + (N + 1) * (y + 1) + (x + 1));
-			}
-	}
-
-	bounds.type = cgltf_primitive_type_triangles;
-	bounds.streams.push_back(bv);
-	bounds.streams.push_back(bc);
 }
 #endif
